! Description:
!> @file
!!   TL of MFASIS fast visible/near-IR scattering model.
!
!> @brief
!!   TL of MFASIS fast visible/near-IR scattering model.
!!
!!
!! @param[out]    err               status on exit
!! @param[in]     chanprof          specifies channels and profiles to simulate
!! @param[in]     chanflag          flags to indicate which channels with LUT available
!! @param[in]     opts              options to configure the simulations
!! @param[in]     profiles          input atmospheric profiles and surface variables
!! @param[in]     profiles_tl       input profile perturbations
!! @param[in]     profiles_int      profiles in internal units
!! @param[in]     profiles_int_tl   profile perturbations in internal units
!! @param[in]     coefs             coefficients structure for instrument to simulate
!! @param[in]     ircld             information on cloud columns
!! @param[in]     ircld_tl          cloud column perturbations
!! @param[in]     aux               additional internal profile variables
!! @param[in]     aux_tl            additional internal profile variable perturbations
!! @param[in]     reflectance       surface BRDFs
!! @param[in]     reflectance_tl    surface BRDF perturbations
!! @param[in]     solar_spectrum    TOA solar irradiance for each channel
!! @param[in]     trans_scatt_ir    cloud/aerosol optical depths
!! @param[in]     trans_scatt_ir_tl cloud/aerosol optical depth perturbations
!! @param[in]     mfasis_refl       quantities computed by rttov_mfasis used by TL/AD/K
!! @param[in,out] radiance_tl       output radiance and BRF perturbations
!
! Copyright:
!    This software was developed within the context of
!    the EUMETSAT Satellite Application Facility on
!    Numerical Weather Prediction (NWP SAF), under the
!    Cooperation Agreement dated 7 December 2016, between
!    EUMETSAT and the Met Office, UK, by one or more partners
!    within the NWP SAF. The partners in the NWP SAF are
!    the Met Office, ECMWF, DWD and MeteoFrance.
!
!    Copyright 2017, EUMETSAT, All Rights Reserved.
!
SUBROUTINE rttov_mfasis_tl( &
              err,              &
              chanprof,         &
              chanflag,         &
              opts,             &
              profiles,         &
              profiles_tl,      &
              profiles_int,     &
              profiles_int_tl,  &
              coefs,            &
              ircld,            &
              ircld_tl,         &
              aux,              &
              aux_tl,           &
              reflectance,      &
              reflectance_tl,   &
              solar_spectrum,   &
              trans_scatt_ir,   &
              trans_scatt_ir_tl,&
              mfasis_refl,      &
              radiance_tl)

#include "throw.h"

  USE parkind1, ONLY : jpim, jprb, jplm

  USE rttov_types, ONLY :        &
    rttov_chanprof,              &
    rttov_options,               &
    rttov_profile,               &
    rttov_coefs,                 &
    rttov_ircld,                 &
    rttov_profile_aux,           &
    rttov_reflectance,           &
    rttov_transmission_scatt_ir, &
    rttov_radiance,              &
    rttov_mfasis_refl
!INTF_OFF
  USE rttov_types, ONLY : &
    rttov_coef_mfasis,           &
    rttov_mfasis_axis

  USE rttov_const, ONLY : &
    wcl_opac_deff,               &
    pi,                          &
    deg2rad,                     &
    pi_r,                        &
    clw_scheme_deff,             &
    mfasis_cld,                  &
!     mfasis_aer,                  &
    mfasis_dim_albedo,           &
    mfasis_dim_kfourier,         &
    mfasis_dim_lfourier,         &
    mfasis_dim_opdp,             &
    mfasis_dim_effdia,           &
    mfasis_dim_scaangle,         &
    gas_id_watervapour,          &
    gas_mass,                    &
    mair,                        &
    gravity,                     &
    nwcl_max

  USE yomhook, ONLY : lhook, dr_hook, jphook
!INTF_ON
  IMPLICIT NONE

  INTEGER(jpim),                      INTENT(OUT)   :: err
  TYPE(rttov_chanprof),               INTENT(IN)    :: chanprof(:)
  LOGICAL(jplm),                      INTENT(IN)    :: chanflag(SIZE(chanprof))
  TYPE(rttov_options),                INTENT(IN)    :: opts
  TYPE(rttov_profile),                INTENT(IN)    :: profiles(:)
  TYPE(rttov_profile),                INTENT(IN)    :: profiles_tl(:)
  TYPE(rttov_profile),                INTENT(IN)    :: profiles_int(:)
  TYPE(rttov_profile),                INTENT(IN)    :: profiles_int_tl(:)
  TYPE(rttov_coefs),                  INTENT(IN)    :: coefs
  TYPE(rttov_ircld),                  INTENT(IN)    :: ircld
  TYPE(rttov_ircld),                  INTENT(IN)    :: ircld_tl
  TYPE(rttov_profile_aux),            INTENT(IN)    :: aux
  TYPE(rttov_profile_aux),            INTENT(IN)    :: aux_tl
  TYPE(rttov_reflectance),            INTENT(IN)    :: reflectance(SIZE(chanprof))
  TYPE(rttov_reflectance),            INTENT(IN)    :: reflectance_tl(SIZE(chanprof))
  REAL(jprb),                         INTENT(IN)    :: solar_spectrum(SIZE(chanprof))
  TYPE(rttov_transmission_scatt_ir),  INTENT(IN)    :: trans_scatt_ir
  TYPE(rttov_transmission_scatt_ir),  INTENT(IN)    :: trans_scatt_ir_tl
  TYPE(rttov_mfasis_refl),            INTENT(IN)    :: mfasis_refl(0:,:)
  TYPE(rttov_radiance),               INTENT(INOUT) :: radiance_tl
!INTF_END

#include "rttov_errorreport.interface"

  INTEGER(jpim)              :: nchanprof, prof, ncolms, npar, chan, lut
  INTEGER(jpim)              :: nlay, nlev, ndim, nid
  INTEGER(jpim)              :: d, n, i, j, k, cc, par, jj
  TYPE(rttov_coef_mfasis)    :: mfasis_coefs
  REAL(jprb),    ALLOCATABLE :: ip   (:)
  REAL(jprb),    ALLOCATABLE :: ip_tl(:), iw_tl(:)

  REAL(jprb),    ALLOCATABLE :: q_mxr(:),dvap(:)
  REAL(jprb),    ALLOCATABLE :: q_mxr_tl(:),dvap_tl(:)
  REAL(jprb)                 :: wvint_bot, wvint_top 
  REAL(jprb)                 :: wvint_bot_tl, wvint_top_tl 
! REAL(jprb)                 :: iw_wv(3)
  REAL(jprb)                 :: iw_wv_tl(3)
  INTEGER(jpim)              :: n_wvdim

  INTEGER(jpim), ALLOCATABLE :: di(:)
  REAL(jprb),    ALLOCATABLE :: od   (:,:), od_eff   (:), ed   (:,:), ed_eff(:), ed_aux(:), od_mpcorr(:,:)
  ! versions of nonlinear quantities needed for TL/AD/K:
  REAL(jprb),    ALLOCATABLE :: od_v1(:,:), od_eff_v1(:), od_mpcorr_v1(:,:)
  REAL(jprb),    ALLOCATABLE :: od_tl(:,:), od_eff_tl(:), ed_tl(:,:), ed_eff_tl(:), ed_eff_v1(:), od_mpcorr_tl(:,:)
  REAL(jprb),    ALLOCATABLE :: odlay(:), odlay_tl(:)

  REAL(jprb)                 :: albedo
  REAL(jprb)                 :: albedo_tl
  REAL(jprb)                 :: theta, theta0, mu, mu0
  TYPE(rttov_mfasis_axis)    :: axis
  REAL(jprb)                 :: colwei_clr, colwei_cc
  REAL(jprb)                 :: refl_tl, rflcol_tl, refl_clr_tl, colwei_tl
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  REAL(jprb)                 :: od1_int
  REAL(jprb)                 :: od1_int_tl, ooo1_tl
  REAL(jprb),    ALLOCATABLE :: ooo1(:)
  REAL(jprb),    ALLOCATABLE :: fac_mx(:)
  REAL(jprb),    ALLOCATABLE :: fac_mx_tl(:)
  REAL(jprb)                 :: odint, a, zenmax, ampl, sigma, delta
  REAL(jprb)                 :: odint_tl
  REAL(jprb),    ALLOCATABLE :: ed_eff_sscat(:), fac_rm(:)
  REAL(jprb),    ALLOCATABLE :: ed_eff_sscat_tl(:), fac_rm_tl(:)
  REAL(jprb),    ALLOCATABLE :: w_sscat(:,:)
  REAL(jprb),    ALLOCATABLE :: w_sscat_tl(:,:)

  REAL(jphook) :: zhook_handle
  ! --------------------------------------------------------------------------------------

  TRY

  IF (LHOOK) CALL DR_HOOK('RTTOV_MFASIS_TL', 0_jpim, ZHOOK_HANDLE)

  ! TODO: Adapt to water vapor correction using three LUTs
  ! TODO: Test aerosol functionality

  ! --------------------------------------------------------------------------
  ! Initialisation
  ! --------------------------------------------------------------------------
  nchanprof = SIZE(chanprof)
  nlay = profiles(1)%nlayers
  nlev = nlay+1

  ! Check whether cloud or aerosol simulation (currently both simultaneously not supported)
  IF (opts%rt_ir%addclouds ) THEN
    mfasis_coefs = coefs%coef_mfasis_cld
  ELSE
    mfasis_coefs = coefs%coef_mfasis_aer
  ENDIF

  npar = mfasis_coefs%nparticles        ! # of particles: 2 for clouds (water and ice)
                                        !                 or # of aerosols considered
  ndim = mfasis_coefs%ndims             ! # of LUT dimensions

  ! Find number of dimensions that are not interpolated (fourier coefs and albedo)
  n = 0_jpim
  DO d = 1, ndim
    IF (mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_kfourier .AND. &
        mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_lfourier .AND. &
        mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_albedo ) CYCLE
    n = n + 1
  END DO

  nid = ndim - n ! Total number of dimensions to be interpolated (total - albedo - fourier indices)

  ALLOCATE( ip     (nid) , &      ! interpolation point
            ip_tl  (nid) , &      ! interpolation point
            iw_tl  (nid) , &      ! interpolation weight
            di     (nid), STAT=err )  ! original dimension index in nid arrays
  THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_tl failed")


  ! Populate di and set some relevant dimensions indices
  di(:) = -1_jpim ! Indices of dimensions to interpolate

  n = 1_jpim
  DO d = 1, ndim
    SELECT CASE (mfasis_coefs%lut_axes(d)%dim_type)
    CASE (mfasis_dim_kfourier, mfasis_dim_lfourier, mfasis_dim_albedo)

    CASE DEFAULT
      di(n) = d
      n = n + 1
    END SELECT
  ENDDO


  ALLOCATE( od_eff    (npar)     , & ! Effective total optical depth per type of particle
            od_eff_v1 (npar)     , & ! version of od_eff for adjoint computations
            od_eff_tl (npar)     , & ! Effective total optical depth per type of particle
            od     (npar, nlay)  , & ! Effective optical depth per type of particle and layer
            od_v1  (npar, nlay)  , & ! od saved for adjoint computations
            od_tl  (npar, nlay)  , & ! Effective optical depth per type of particle and layer
            od_mpcorr(npar, nlay), &   ! Effective mixed-phase corrected optical depth per type of particle and layer
            od_mpcorr_v1(npar, nlay), &! od_mpcorr saved for adjoint computations
            od_mpcorr_tl(npar, nlay), &! Effective mixed-phase corrected optical depth per type of particle and layer
            odlay   (nlay)          , &   ! Effective mixed-phase corrected optical depth of water and ice
            odlay_tl(nlay)          , &   ! Effective mixed-phase corrected optical depth of water and ice
            STAT=err)
  THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_tl failed")

  ! For the time being we assume effective diameters will be out of LUT for aerosols
  IF ( mfasis_coefs%file_type .EQ. 1 ) THEN
    ALLOCATE( ed_eff (npar)      , & ! Total effective diameter per type of particle
              ed_aux (npar)      , & ! Effective diameters to be used with 0 opdp to avoid flagging
              ed_eff_v1 (npar)   , & ! Total effective diameter per type of particle
              ed_eff_tl(npar)    , & ! Total effective diameter per type of particle
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              fac_mx (nlay),      &         ! fraction of cloud considered as (potentially) mixed cloud
              fac_mx_tl (nlay),   &         ! fraction of cloud considered as (potentially) mixed cloud
              ooo1 (nlay),        &         ! discriminator for switching on mixed layer summation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              ed   (npar, nlay)  , & ! Effective diameter per type of particle and layer
              ed_tl(npar, nlay),   &
              w_sscat(npar, nlay), &     ! Weighting factor for radius mixing approach (1.6um channel)
              w_sscat_tl(npar, nlay), &  ! Weighting factor for radius mixing approach (1.6um channel)
              ed_eff_sscat(npar), &      ! Effective diameter contribution in radius mixing approach (1.6um channel)
              ed_eff_sscat_tl(npar),  &  ! Effective diameter contribution in radius mixing approach (1.6um channel)
              fac_rm(npar),    &         ! Weighting factor for radius mixing approach (1.6um channel)
              fac_rm_tl(npar), &         ! Weighting factor for radius mixing approach (1.6um channel)
              q_mxr(nlev),         &
              q_mxr_tl(nlev),      &
              dvap (nlay),         &
              dvap_tl (nlay),      &
              STAT=err) !
    THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_tl failed")

    ! Effective diameters to be used with 0 opdp to avoid incorrect flagging for clear column
    n = 1_jpim
    DO d = 1, ndim
      IF (mfasis_coefs%lut_axes(d)%dim_type .NE. mfasis_dim_effdia) CYCLE
      ed_aux(n) = mfasis_coefs%lut_axes(d)%values(1)
      n = n + 1
    ENDDO
  ENDIF
  fac_mx (:) = 0
  ooo1 (:)   = 0
  ! --------------------------------------------------------------------------
  ! Channel loop
  ! --------------------------------------------------------------------------
  DO i = 1, nchanprof
    IF (.NOT. chanflag(i)) CYCLE
    prof = chanprof(i)%prof
    chan = chanprof(i)%chan
    lut = mfasis_coefs%channel_lut_index(chan)

    n_wvdim=size(mfasis_coefs%lut(lut)%qint,2)
    dvap(:)=0._jprb
    dvap_tl(:)=0._jprb
    IF(n_wvdim==3) THEN
      q_mxr(:) =  profiles_int(prof)%q(:)* gas_mass(gas_id_watervapour)/  &
                  (mair * 1.E06_jprb +profiles_int(prof)%q(:)*gas_mass(gas_id_watervapour))

      q_mxr_tl(:)= profiles_int_tl(prof)%q(:)*gas_mass(gas_id_watervapour) * (1._jprb - q_mxr(:) ) &
                  /(mair * 1.E06_jprb +profiles_int(prof)%q(:)*gas_mass(gas_id_watervapour))

      dvap(1:nlay)  = 100._jprb * (profiles(prof)%p(2:nlev)-profiles(prof)%p(1:nlev-1))* &
                                           0.5_jprb * (q_mxr(2:nlev)+q_mxr(1:nlev-1))/gravity


       IF (opts%interpolation%lgradp) THEN
         dvap_tl(1:nlay)=100._jprb *((profiles_tl(prof)%p(2:nlev)-profiles_tl(prof)%p(1:nlev-1))* &
                                              0.5_jprb * (q_mxr(2:nlev)+q_mxr(1:nlev-1))/gravity  &
                                     + (profiles(prof)%p(2:nlev)-profiles(prof)%p(1:nlev-1))* &
                                        0.5_jprb * (q_mxr_tl(2:nlev)+q_mxr_tl(1:nlev-1))/gravity  &
                                    )
      ELSE
        dvap_tl(1:nlay)=100._jprb *((profiles(prof)%p(2:nlev)-profiles(prof)%p(1:nlev-1))* &
                                        0.5_jprb * (q_mxr_tl(2:nlev)+q_mxr_tl(1:nlev-1))/gravity  &
                                    )
      ENDIF
 
      jj = aux%s(prof)%nearestlev_surf - 1
      IF(jj>=1 .AND. jj<=nlay) THEN
        dvap_tl(jj)=dvap_tl(jj)*(1._jprb - aux   %s(prof)%pfraction_surf)           &
                        - dvap   (jj) * aux_tl%s(prof)%pfraction_surf
        dvap   (jj)=dvap   (jj)*(1._jprb - aux   %s(prof)%pfraction_surf)
      ENDIF
    ENDIF

    colwei_clr = ircld%xcolclr(prof) ! clear column
    ! --------------------------------------------------------------------------
    ! Albedo
    ! --------------------------------------------------------------------------
    albedo   =     reflectance(i)%refl_out * pi
    IF(albedo < 1._jprb) THEN
      albedo_tl= reflectance_tl(i)%refl_out * pi
    ELSE
      albedo   = 1._jprb
      albedo_tl= 0._jprb
    ENDIF

    ! --------------------------------------------------------------------------
    ! Angles (in radians except alpha_deg)
    ! --------------------------------------------------------------------------
    theta = profiles(prof)%zenangle * deg2rad
    theta0 = profiles(prof)%sunzenangle * deg2rad
    mu  = COS(theta)
    mu0 = COS(theta0)

    ip   (:) = -1._jprb          ! interpolation point
    ip_tl(:) =  0._jprb          ! interpolation point


!   iw   (:)  = 0._jprb  ! interpolation weight
    iw_tl(:)  = 0._jprb  ! interpolation weight


    ! --------------------------------------------------------------------------
    ! Loop over (cloud) columns
    ! --------------------------------------------------------------------------
    ! No cloud columns for aerosols calculations
    IF ( mfasis_coefs%file_type == mfasis_cld ) THEN
      ncolms = ircld%ncolumn(prof)
    ELSE
      ncolms = 1_jpim
    ENDIF


    refl_tl     = 0._jprb
    refl_clr_tl = 0._jprb

    DO cc = 1, ncolms
!==================================================================================
! First compute nonlinear quatities needed in linear computations (for given channel "i" and column "cc")
!==================================================================================
      ! --------------------------------------------------------------------------
      ! Set column weight, optical depth and effective diameters
      ! --------------------------------------------------------------------------
      IF ( mfasis_coefs%file_type .EQ. mfasis_cld ) THEN ! clouds
        colwei_cc = ircld%xcol(cc+1,prof) - ircld%xcol(cc,prof)
        od   (:,:) = 0._jprb
        ed   (:,:) = 0._jprb
        od1_int = 0._jprb
        od_mpcorr(:,:) = 0._jprb
        odint = 0._jprb
        fac_mx(:)= 0._jprb
        wvint_bot=0._jprb
        wvint_top=0._jprb

        DO j = 1, nlay
          IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
          IF ( ircld%icldarr(cc,j,prof) .EQ. 1 ) THEN
            od   (1,j) = SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) ! water clouds
            od   (2,j) = trans_scatt_ir%opdpext(nwcl_max+1,j,i)        ! ice clouds
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              od_v1(:,j) = od   (:,j)
              ! Modify optical depth in partial layer above surface
              od   (:,j) = od   (:,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF

            !-----------------------------------------------------------------------
            ! mixed-phase correction per layer: add ice optical depth below water
            ! cloud top to water optical depth and subtract from ice optical depth
            !-----------------------------------------------------------------------
            od1_int = od1_int + od(1,j)
            ooo1(j)=od1_int-opts%dev%od1_thresh
            fac_mx(j)=(1._jprb + TANH(ooo1(j)/opts%dev%o_del1))/2._jprb

            od_mpcorr(1,j) = od(1,j) + fac_mx(j)*od(2,j)
            od_mpcorr(2,j) = od(2,j) - fac_mx(j)*od(2,j)
            od_mpcorr_v1(1,j) =  od_mpcorr(1,j)
            od_mpcorr_v1(2,j) =  od_mpcorr(2,j)
            IF (od_mpcorr_v1(2,j) < 0) od_mpcorr(2,j) = 0._jprb

            !-----------------------------------------------------------------------
            ! determine effective diameter
            !-----------------------------------------------------------------------
            IF ( coefs%coef_mfasis_cld%clw_scheme == clw_scheme_deff ) THEN ! Deff water clouds scheme
              ed   (1,j) = aux%clw_dg(j,prof)
            ELSE ! OPAC water cloud parametrization
              ed   (1,j) = SUM(wcl_opac_deff(1:nwcl_max) * trans_scatt_ir%opdpext(1:nwcl_max,j,i))
              IF ( od(1,j) .GT. 0. ) THEN
                ! TODO: take into account optical depth surface correction in eff diameter calculation
                ed   (1,j) =  ed   (1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
              ELSE
                ed   (1,j) = 0._jprb
              ENDIF
            ENDIF
            ed   (2,j) = aux%ice_dg(j,prof)
          ELSEIF(j>1) THEN
            fac_mx(j) = fac_mx(j-1)
          ENDIF
          IF(n_wvdim==3) THEN
            wvint_bot = wvint_bot + fac_mx(j)          * dvap(j)
            wvint_top = wvint_top + (1._jprb-fac_mx(j))* dvap(j)
          ENDIF
        ENDDO
        !-----------------------------------------------------------------------
        ! determine integrated optical depths and effective diameter
        !-----------------------------------------------------------------------
        ! water: use original optical depths for computing the effective diameter, then overwrite od_eff(1)
        od_eff(1) = SUM(od(1,:))
        od_eff_v1(1) = od_eff(1)
        IF ( od_eff_v1(1) .GT. 0. ) THEN
          ed_eff(1) = SUM( od(1,:)*ed(1,:))/od_eff_v1(1)
          ed_eff_v1(1) = ed_eff(1)
        ELSE
          ed_eff(1) = ed_aux(1)
          ed_eff_v1(1) = ed_eff(1)
        ENDIF
        od_eff(1) = SUM(od_mpcorr(1,:))

        ! ice: use mixed-phase corrected optical depths for computing the effective diameter
        od_eff(2) = SUM(od_mpcorr(2,:))
        od_eff_v1(2) = od_eff(2)
        IF ( od_eff_v1(2) .GT. 0. ) THEN
          ed_eff(2) = SUM( od_mpcorr(2,:)*ed(2,:))/od_eff_v1(2)
          ed_eff_v1(2) = ed_eff(2)
        ELSE
          ed_eff(2) = ed_aux(2)
          ed_eff_v1(2) = ed_eff(2)
        ENDIF

        !-----------------------------------------------------------------------
        ! special treatment for nir channels: radius mixing approach
        ! (currently only for 1.6um channel)
        !-----------------------------------------------------------------------
        IF (mfasis_coefs%channel_deff_mixing(lut) .EQ. 1) THEN
          ! compute weights for computing single-scattering effective diameter
          ! (scattering takes place mainly in upper cloud layers)
          ! R = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * exp(-(od(1,z)+od(2,z))/mu) * exp(-(od(1,z)+od(2,z))/mu0) dz
          !   = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * exp(-(od(1,z)+od(2,z))/a) dz
          ! note: since exp(-(od(1,z)+od(2,z))/a) is nonlinear, use mean of function for integrating
          ! 1/(tau_j - tau_{j-1}) * int_0^{tau_j - tau_{j-1}} exp(-(x+tau_{j-1})/a) dx = 
          !     = a/(tau_j - tau_{j-1}) * ( exp((tau_j - tau_{j-1})/a) - 1) * exp(-tau_j/a)
          !     = a/(od(1,j) + od(2,j)) * ( exp((od(1,j) + od(2,j) - odint)/a) - exp(-odint/a) )
          ! -> R = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * (a*/(od(1,j) + od(2,j))) * &
          !        ( exp((od(1,j) + od(2,j) - odint)/a) - exp(-odint/a) ) * dz
          ! use beta(k,z) * dz = od(k,j)/dz * dz = od(k,j)
          
          a = ( mu*mu0 ) / ( mu+mu0 ) ! factor from integral
          w_sscat(:,:) = 0._jprb
          ed_eff_sscat(:) = 0._jprb
          odint = 0._jprb
          
          ! ice phase (neglect water phase for the moment, has only little effect)
          IF (od_eff(2) > 0.1_jprb) THEN ! exclude special treatment for very thin clouds
            DO j = 1, nlay
              IF ( ircld%icldarr(cc,j,prof) .EQ. 1 ) THEN ! this check is probably redundant, od_eff > 0!
                odlay(j) = od_mpcorr(1,j) + od_mpcorr(2,j)
                odint = odint + odlay(j) ! total integraged optical depth (mpc corr.)
                IF ( odlay(j) > 1E-10 .and. a > 1E-10 ) THEN
                  w_sscat(2,j) = (a*od_mpcorr(2,j)/odlay(j)) * ( EXP((odlay(j) - odint)/a) - EXP(-odint/a) )
                ENDIF
              ENDIF
            ENDDO
          ENDIF
          IF (sum(w_sscat(2,:)) > 0) THEN
              ed_eff_sscat(2) = SUM( ed(2,:) * w_sscat(2,:)) / SUM(w_sscat(2,:))
          ENDIF

          !-----------------------------------------------------------------------
          ! Radius mixing parameterized with Gaussian function:
          zenmax = MAX(theta,theta0)
          fac_rm(:) = 0._jprb
          ! Ice phase:
          ampl = 0.6790 - 0.7840*zenmax + 1.1925*zenmax**2
          sigma = 0.4927 - 0.2109*zenmax + 0.4710*zenmax**2
          delta = 1E-6 ! avoid numerical problems for od_eff(2) ~ 0
          fac_rm(2) = ampl/(sigma*SQRT(2._jprb*pi)) * DEXP(-0.5*((LOG(od_eff(2)+delta) - 3.15)/sigma)**2)
          
          ed_eff(2) = fac_rm(2)*ed_eff_sscat(2) + (1-fac_rm(2))*ed_eff(2)

        ENDIF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      ELSE  ! aerosols
        colwei_cc = 1._jprb     ! single column for aerosols
!       colwei_tl = 0._jprb     ! single column for aerosols
        od   (:,:)= 0._jprb
!       od_tl(:,:)= 0._jprb
        DO k = 1, npar
          par = mfasis_coefs%aer_types(k)
          DO j = 1, nlay
            IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
            od(k,j)    = trans_scatt_ir%opdpext(par,j,i)
!           od_tl(k,j) = trans_scatt_ir_tl%opdpext(par,j,i)
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              od_v1(:,j) = od   (:,j)
              ! Modify optical depth in partial layer above surface
!             od_tl(k,j) = od_tl(k,j) * (1 - aux%s(prof)%pfraction_surf)  &
!                         - od_v1(k,j) * aux_tl%s(prof)%pfraction_surf
              od   (k,j) = od   (k,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF
          ENDDO
          od_eff   (k) = SUM(od   (k,:))
!         od_eff_tl(k) = SUM(od_tl(k,:))
        ENDDO
      ENDIF
      ! --------------------------------------------------------------------------
      ! Fill into the interpolation point array the rest of dimensions
      ! --------------------------------------------------------------------------
      j = 1
      k = 1
      DO d = 1, nid
        SELECT CASE (mfasis_coefs%lut_axes(di(d))%dim_type)
        CASE (mfasis_dim_opdp)
          ip   (d) = od_eff   (j)
          j = j + 1
        CASE (mfasis_dim_effdia)
          ! Check IF stored as radius or diamater in LUT
          IF (mfasis_coefs%lut_axes(di(d))%name(1:1) .EQ. "R") THEN
            ip(d) = ed_eff(k)/2
          ELSE
            ip(d) = ed_eff(k)
          ENDIF
          k = k + 1
        END SELECT
      ENDDO


!==================================================================================
! Start linear computations for channel "i" and column "cc"
!==================================================================================
      ! --------------------------------------------------------------------------
      ! Set column weight, optical depth and effective diameters
      ! --------------------------------------------------------------------------
      IF ( mfasis_coefs%file_type .EQ. mfasis_cld ) THEN ! clouds
        colwei_tl = ircld_tl%xcol(cc+1,prof) - ircld_tl%xcol(cc,prof)
        od_tl(:,:) = 0._jprb
        ed_tl(:,:) = 0._jprb

        od1_int_tl = 0._jprb
        od_mpcorr_tl(:,:) = 0._jprb
        fac_mx_tl(:)= 0._jprb
        wvint_bot_tl=0._jprb
        wvint_top_tl=0._jprb

        DO j = 1, nlay
          IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
          IF ( ircld%icldarr(cc,j,prof) .EQ. 1 ) THEN
            od_tl(1,j) = SUM(trans_scatt_ir_tl%opdpext(1:nwcl_max,j,i)) ! water clouds
            od_tl(2,j) = trans_scatt_ir_tl%opdpext(nwcl_max+1,j,i)        ! ice clouds
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
              od_tl(:,j) = od_tl(:,j) * (1 - aux%s(prof)%pfraction_surf)  &
                               - od_v1(:,j) * aux_tl%s(prof)%pfraction_surf
            ENDIF
            !-----------------------------------------------------------------------
            ! mixed-phase correction per layer: add ice optical depth below water
            ! cloud top to water optical depth and subtract from ice optical depth
            !-----------------------------------------------------------------------
            od1_int_tl = od1_int_tl + od_tl(1,j)
            ooo1_tl=od1_int_tl
            fac_mx_tl(j)= 0._jprb
            IF(abs(ooo1(j)/opts%dev%o_del1) < 5._jprb) &
               fac_mx_tl(j)=(ooo1_tl/opts%dev%o_del1) / (COSH(ooo1(j)/opts%dev%o_del1)**2) /2._jprb

             od_mpcorr_tl(1,j) = od_tl(1,j) + fac_mx_tl(j)*od(2,j) + fac_mx(j)*od_tl(2,j)
             od_mpcorr_tl(2,j) = od_tl(2,j) - fac_mx_tl(j)*od(2,j) - fac_mx(j)*od_tl(2,j)
            IF (od_mpcorr_v1(2,j) < 0) od_mpcorr_tl(2,j) = 0._jprb

            !-----------------------------------------------------------------------
            ! determine effective diameter
            !-----------------------------------------------------------------------
            IF ( coefs%coef_mfasis_cld%clw_scheme == clw_scheme_deff ) THEN ! Deff water clouds scheme
              ed_tl(1,j) = aux_tl%clw_dg(j,prof)
            ELSE ! OPAC water cloud parametrization
              ed_tl(1,j) = SUM(wcl_opac_deff(1:nwcl_max) * trans_scatt_ir_tl%opdpext(1:nwcl_max,j,i))
              IF ( od(1,j) .GT. 0. ) THEN
                ! TODO: take into account optical depth surface correction in eff diameter calculation
                ed_tl(1,j) =  ed_tl(1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) - &
                              ed(1,j)*SUM(trans_scatt_ir_tl%opdpext(1:nwcl_max,j,i))/&
                              SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
              ELSE
                ed_tl(1,j) = 0._jprb
              ENDIF
            ENDIF
            ed_tl(2,j) = aux_tl%ice_dg(j,prof)
          ELSEIF(j>1) THEN
            fac_mx_tl(j) = fac_mx_tl(j-1)
          ENDIF
          IF(n_wvdim==3) THEN
              wvint_bot_tl = wvint_bot_tl + fac_mx(j)          * dvap_tl(j) + fac_mx_tl(j) * dvap(j)
              wvint_top_tl = wvint_top_tl + (1._jprb-fac_mx(j))* dvap_tl(j) - fac_mx_tl(j) * dvap(j)
          ENDIF
        ENDDO

        !-----------------------------------------------------------------------
        ! determine integrated optical depths and effective diameter
        !-----------------------------------------------------------------------
        ! water: use original optical depths for computing the effective diameter, then overwrite od_eff(1)
        od_eff_tl(1) = SUM( od_tl(1,:))
        IF ( od_eff_v1(1) .GT. 0. ) THEN
          ed_eff_tl(1) = SUM( od_tl(1,:)*ed(1,:) + od(1,:)*ed_tl(1,:))/od_eff_v1(1)   &
                           - ed_eff_v1(1) * od_eff_tl(1)/od_eff_v1(1)
        ELSE
          ed_eff_tl(1) = 0._jprb
        ENDIF
        od_eff_tl(1) = SUM( od_mpcorr_tl(1,:) )

        ! ice: use mixed-phase corrected optical depths for computing the effective diameter
        od_eff_tl(2) = SUM( od_mpcorr_tl(2,:))
        IF ( od_eff_v1(2) .GT. 0. ) THEN
          ed_eff_tl(2) = SUM( od_mpcorr_tl(2,:)*ed(2,:) + od_mpcorr(2,:)*ed_tl(2,:))/od_eff_v1(2) &
                           - ed_eff_v1(2) * od_eff_tl(2)/od_eff_v1(2)
        ELSE
          ed_eff_tl(2) = 0._jprb
        ENDIF


        !-----------------------------------------------------------------------
        ! special treatment for nir channels: radius mixing approach
        ! (currently only for 1.6um channel)
        !-----------------------------------------------------------------------
        IF (mfasis_coefs%channel_deff_mixing(lut) .EQ. 1) THEN
          ! compute weights for computing single-scattering effective diameter 
          ! (scattering takes place mainly in upper cloud layers)
          ! R = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * exp(-(od(1,z)+od(2,z))/mu) * exp(-(od(1,z)+od(2,z))/mu0) dz
          !   = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * exp(-(od(1,z)+od(2,z))/a) dz
          ! note: since exp(-(od(1,z)+od(2,z))/a) is nonlinear, use mean of function for integrating
          ! 1/(tau_j - tau_{j-1}) * int_0^{tau_j - tau_{j-1}} exp(-(x+tau_{j-1})/a) dx = 
          !     = a/(tau_j - tau_{j-1}) * ( exp((tau_j - tau_{j-1})/a) - 1) * exp(-tau_j/a)
          !     = a/(od(1,j) + od(2,j)) * ( exp((od(1,j) + od(2,j) - odint)/a) - exp(-odint/a) )
          ! -> R = N^{-1} \int_0^{TOA} r(k,z) * beta(k,z) * (a*/(od(1,j) + od(2,j))) * &
          !        ( exp((od(1,j) + od(2,j) - odint)/a) - exp(-odint/a) ) * dz
          ! use beta(k,z) * dz = od(k,j)/dz * dz = od(k,j)

          w_sscat_tl(:,:) = 0._jprb
          ed_eff_sscat_tl(:) = 0._jprb
          odint_tl = 0._jprb
          
          ! ice phase (neglect water phase for the moment, has only little effect)
          IF (od_eff(2) > 0.1_jprb) THEN ! exclude special treatment for very thin clouds
            DO j = 1, nlay
              IF ( ircld%icldarr(cc,j,prof) .EQ. 1 ) THEN ! this check is probably redundant, od_eff > 0!
                odlay_tl(j) = od_mpcorr_tl(1,j) + od_mpcorr_tl(2,j)
                odint_tl = odint_tl + odlay_tl(j) ! total integraged optical depth (mpc corr.)
                IF ( odlay(j) > 1E-10 .and. a > 1E-10 ) THEN
                  w_sscat_tl(2,j) = a/odlay(j) * (EXP((odlay(j) - odint)/a) - EXP(-odint/a)) * od_mpcorr_tl(2,j) &
                                  + ( - a* od_mpcorr(2,j)/(odlay(j)**2) * (EXP((odlay(j) - odint)/a) - EXP(-odint/a)) &
                                      + od_mpcorr(2,j)/odlay(j) * EXP((odlay(j) - odint)/a) ) * odlay_tl(j) &
                                  + od_mpcorr(2,j)/odlay(j) * ( -EXP((odlay(j) - odint)/a) + EXP(-odint/a)) * odint_tl
                ENDIF
              ENDIF
            ENDDO
          ENDIF
          IF (sum(w_sscat(2,:)) > 0) THEN
            ed_eff_sscat_tl(2) = SUM( ed_tl(2,:) * w_sscat(2,:))/ SUM(w_sscat(2,:)) &
                                + 1/SUM(w_sscat(2,:)) *  SUM( (ed(2,:)-ed_eff_sscat(2)) * w_sscat_tl(2,:))
          ENDIF

          !-----------------------------------------------------------------------
          ! Radius mixing parameterized with Gaussian function:
          fac_rm_tl(:) = 0._jprb
          ! Ice phase:
          fac_rm_tl(2) = - ampl/(sigma*SQRT(2._jprb*pi)) * DEXP(-0.5*((LOG(od_eff(2)+delta) - 3.15)/sigma)**2) &
                          * (LOG(od_eff(2)+delta) - 3.15)/sigma * 1/sigma * 1/(od_eff(2)+delta) * od_eff_tl(2)
          
          ed_eff_tl(2) = (ed_eff_sscat(2) - ed_eff_v1(2)) * fac_rm_tl(2) &
                       + fac_rm(2) * ed_eff_sscat_tl(2) &
                       + (1 - fac_rm(2)) * ed_eff_tl(2)
        ENDIF

      ELSE  ! aerosols
!       colwei_cc = 1._jprb     ! single column for aerosols
        colwei_tl = 0._jprb     ! single column for aerosols
!       od   (:,:)= 0._jprb
        od_tl(:,:)= 0._jprb
        DO k = 1, npar
          par = mfasis_coefs%aer_types(k)
          DO j = 1, nlay
            IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
!           od(k,j)    = trans_scatt_ir%opdpext(par,j,i)
            od_tl(k,j) = trans_scatt_ir_tl%opdpext(par,j,i)
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
!             od_v1(k,j) = od   (k,j)
              od_tl(k,j) = od_tl(k,j) * (1 - aux%s(prof)%pfraction_surf)  &
                          - od_v1(k,j) * aux_tl%s(prof)%pfraction_surf
!             od   (k,j) = od   (k,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF
          ENDDO
!         od_eff   (k) = SUM(od   (k,:))
          od_eff_tl(k) = SUM(od_tl(k,:))
        ENDDO
      ENDIF !IF CLOUD ELSE AEROSOL
      ! --------------------------------------------------------------------------
      ! Fill into the interpolation point array the rest of dimensions
      ! --------------------------------------------------------------------------
      j = 1
      k = 1
      DO d = 1, nid
        SELECT CASE (mfasis_coefs%lut_axes(di(d))%dim_type)
        CASE (mfasis_dim_opdp)
          ip_tl(d) = od_eff_tl(j)
          j = j + 1
        CASE (mfasis_dim_effdia)
          ! Check IF stored as radius or diamater in LUT
          IF (mfasis_coefs%lut_axes(di(d))%name(1:1) .EQ. "R") THEN
            ip_tl(d) = ed_eff_tl(k)/2
          ELSE
            ip_tl(d) = ed_eff_tl(k)
          ENDIF
          k = k + 1
        END SELECT
      ENDDO

      ! --------------------------------------------------------------------------
      ! Look for nearest entries in LUT and compute interpolation weights
      ! --------------------------------------------------------------------------
      ! Here we are assuming the LUT dimensions for optical depths (and effective
      ! diameter in the case of clouds) are always ordered as aer_types in the case
      ! of aerosols and first water and THEN ice in the case of clouds
      DO d = 1, nid
        IF (mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle) CYCLE ! Skip scattering angle (already DOne)
        axis = mfasis_coefs%lut_axes(di(d))
        ! Reset lc_aux and iw to make sure correct for every column calculation
!       iw(d)     = 0._jprb  ! interpolation weight
        iw_tl(d)  = 0._jprb

        IF ( ip(d) .GE. axis%values(axis%nvalues) ) THEN
!         iw   (d)  = 1._jprb
          iw_tl(d)  = 0._jprb
        ELSE
          DO j = 1, axis%nvalues - 1
            IF ( ip(d) .GT. axis%values(j) ) THEN
              IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_opdp .AND. j .NE. 1 ) THEN
                ! optical depths ( lin. int. in log)
                iw_tl(d) =  ip_tl(d)/ip(d)                   &
                  / ( LOG(axis%values(j+1)) - LOG(axis%values(j)) )
              ELSE            ! effective radii and alpha, linear interpolation
                iw_tl(d) =  ip_tl(d)               &
                  / ( axis%values(j+1) - axis%values(j) )
              ENDIF
            ENDIF
          ENDDO
        ENDIF

      ENDDO
      iw_wv_tl(:) = 0._jprb
      if( n_wvdim==3) THEN
          iw_wv_tl(2)=                  wvint_bot_tl  / &
               (mfasis_coefs%lut(lut)%qint(1,2) - mfasis_coefs%lut(lut)%qint(1,1))
          iw_wv_tl(3)=                  wvint_top_tl  / &
               (mfasis_coefs%lut(lut)%qint(2,3) - mfasis_coefs%lut(lut)%qint(2,1))
          iw_wv_tl(1) =       - iw_wv_tl(2) - iw_wv_tl(3)
       ENDIF


      ! --------------------------------------------------------------------------
      ! Linear interpolation in nid dimensions (replaces subroutine rttov_mfasis_interpolate)
      ! --------------------------------------------------------------------------

      rflcol_tl=0
      DO d = 1, nid
        IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle ) THEN
          rflcol_tl = rflcol_tl+mfasis_refl(cc,i)%refl_lin_coef(d)*albedo_tl
        ELSE
          rflcol_tl = rflcol_tl+mfasis_refl(cc,i)%refl_lin_coef(d)*iw_tl(d)
        ENDIF
      ENDDO
      rflcol_tl = rflcol_tl+sum(iw_wv_tl(1:n_wvdim)*mfasis_refl(cc,i)%refl_wv(1:n_wvdim))


      ! --------------------------------------------------------------------------
      ! Add column contribution to total reflectance
      ! --------------------------------------------------------------------------

      refl_tl = refl_tl + rflcol_tl*colwei_cc + mfasis_refl(cc,i)%refl*colwei_tl

    ENDDO ! END of cloud column loop

    ! --------------------------------------------------------------------------
    ! Compute clear reflectance and add clear column to total reflectance
    ! --------------------------------------------------------------------------
    ! Rerun linear interpolation in nid dimensions setting optical depths
    ! to zero with weight 1
    ! This assumes zero optical depth is always the first value
    IF(n_wvdim==3) THEN
      wvint_bot_tl=0._jprb
      wvint_top_tl=0._jprb
      DO j = 1, nlay
        wvint_top_tl = wvint_top_tl +  dvap_tl(j)
      ENDDO
    ENDIF
    iw_wv_tl(:) = 0._jprb
    if( n_wvdim==3) THEN
        iw_wv_tl(2)=                  wvint_bot_tl  / &
             (mfasis_coefs%lut(lut)%qint(1,2) - mfasis_coefs%lut(lut)%qint(1,1))
        iw_wv_tl(3)=                  wvint_top_tl  / &
             (mfasis_coefs%lut(lut)%qint(2,3) - mfasis_coefs%lut(lut)%qint(2,1))
        iw_wv_tl(1) =       - iw_wv_tl(2) - iw_wv_tl(3)
     ENDIF

    DO d = 1, nid
      IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_opdp ) THEN
!       iw   (d) = 1._jprb
        iw_tl(d) = 0._jprb
      ENDIF
    ENDDO

    refl_clr_tl=0
    DO d = 1, nid
      IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle ) THEN
        refl_clr_tl = refl_clr_tl+mfasis_refl(0,i)%refl_lin_coef(d)*albedo_tl
      ELSE
        refl_clr_tl = refl_clr_tl+mfasis_refl(0,i)%refl_lin_coef(d)*iw_tl(d)
      ENDIF
    ENDDO
    refl_clr_tl = refl_clr_tl+sum(iw_wv_tl(1:n_wvdim)*mfasis_refl(0,i)%refl_wv(1:n_wvdim))

    colwei_tl = ircld_tl%xcolclr(prof) ! clear column

    refl_tl = refl_tl + refl_clr_tl*colwei_clr + mfasis_refl(0,i)%refl*colwei_tl

    ! --------------------------------------------------------------------------
    ! Fill radiance structure
    ! --------------------------------------------------------------------------
    radiance_tl%refl(i) = refl_tl
    radiance_tl%refl_clear(i) = refl_clr_tl

    radiance_tl%total(i) = refl_tl * solar_spectrum(i) * pi_r  &
      * COS(profiles(prof)%sunzenangle * deg2rad)
    radiance_tl%clear(i) = refl_clr_tl * solar_spectrum(i) * pi_r  &
      * COS(profiles(prof)%sunzenangle * deg2rad)
    radiance_tl%cloudy(i) = radiance_tl%total(i)

  ENDDO ! END of channel loop

  ! --------------------------------------------------------------------------
  ! Tidy up
  ! --------------------------------------------------------------------------

  DEALLOCATE( ip, ip_tl, iw_tl, di, od, od_tl, od_v1, od_eff, od_eff_tl, od_eff_v1, &
              od_mpcorr, od_mpcorr_v1, od_mpcorr_tl, odlay, odlay_tl, STAT=err)
  THROWM(err.NE.0,"Deallocation of memory for rttov_mfasis_tl failed")
  IF (mfasis_coefs%file_type .EQ. mfasis_cld) THEN
    DEALLOCATE(ed, ed_tl, ed_eff, ed_aux, ed_eff_tl, ed_eff_v1, fac_mx, fac_mx_tl, &
               q_mxr, q_mxr_tl, dvap, dvap_tl, STAT=err)
    THROWM(err.NE.0,"Deallocation of memory for rttov_mfasis_tl failed")
    DEALLOCATE(w_sscat, w_sscat_tl, fac_rm, fac_rm_tl, ed_eff_sscat, ed_eff_sscat_tl, STAT=err)
    THROWM(err.NE.0,"Deallocation of memory for rttov_mfasis_tl failed")
  ENDIF

  IF (LHOOK) CALL DR_HOOK('RTTOV_MFASIS_TL', 1_jpim, ZHOOK_HANDLE)

  CATCH

  IF (LHOOK) CALL DR_HOOK('RTTOV_MFASIS_TL', 1_jpim, ZHOOK_HANDLE)

END SUBROUTINE rttov_mfasis_tl
